/*****************************************************************
 
  e_subtext.c
 
  Copyright (c) 2006 by Burkhard Plaum - plaum@ipf.uni-stuttgart.de
 
  http://gmerlin.sourceforge.net
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 
*****************************************************************/

#include <stdio.h>
#include <string.h>

#include <plugin.h>
#include <utils.h>

typedef struct
  {
  FILE * output;  
  int format_index;
  char * filename;
  int titles_written;
  } subtext_t;

static void write_time_srt(FILE * output, gavl_time_t time)
  {
  int msec, sec, m, h;

  msec = (time % GAVL_TIME_SCALE) / (GAVL_TIME_SCALE/1000);
  time /= GAVL_TIME_SCALE;
  sec = time % 60;
  time /= 60;
  m = time % 60;
  time /= 60;
  h = time;

  fprintf(output, "%02d:%02d:%02d,%03d", h, m, sec, msec);
  }

static void write_subtitle_srt(subtext_t * s, const char * text,
                               gavl_time_t time, gavl_time_t duration)
  {
  int i;
  char ** lines;
  /* Index */

  fprintf(s->output, "%d\r\n", s->titles_written + 1);
  
  /* Time */
  write_time_srt(s->output, time);
  fprintf(s->output, " --> ");
  write_time_srt(s->output, time+duration);
  fprintf(s->output, "\r\n");
  
  lines = bg_strbreak(text, '\n');
  i = 0;
  while(lines[i])
    {
    fprintf(s->output, "%s\r\n", lines[i]);
    i++;
    }
  /* Empty line */
  fprintf(s->output, "\r\n");
  bg_strbreak_free(lines);
  }

static void write_subtitle_mpsub(subtext_t * s, const char * text,
                                 gavl_time_t time, gavl_time_t duration)
  {
  
  }

static struct
  {
  const char * extension;
  const char * name;
  void (*write_subtitle)(subtext_t * s, const char * text, gavl_time_t time, gavl_time_t duration);
  }
formats[] =
  {
    {
      extension:      "srt",
      name:           "srt",
      write_subtitle: write_subtitle_srt,
    },
    {
      extension:      "sub",
      name:           "mpsub",
      write_subtitle: write_subtitle_mpsub,
    }
  };

static void * create_subtext()
  {
  subtext_t * ret = calloc(1, sizeof(*ret));
  return ret;
  }

static const char * get_extension_subtext(void * data)
  {
  subtext_t * e;
  e = (subtext_t *)data;
  return formats[e->format_index].extension;
  }

static int open_subtext(void * data, const char * filename,
                        bg_metadata_t * metadata)
  {
  subtext_t * e;
  e = (subtext_t *)data;

  e->filename = bg_strdup(e->filename, filename);
  e->output = fopen(e->filename, "w");
  
  return 1;
  }

static int add_subtitle_text_stream_subtext(void * data, const char * language)
  {
  subtext_t * e;
  e = (subtext_t *)data;
  return 0;
  }

static int start_subtext(void * data)
  {
  return 1;
  }

static void write_subtitle_text_subtext(void * data, const char * text, gavl_time_t start,
                                        gavl_time_t duration, int stream)
  {
  subtext_t * e;
  e = (subtext_t *)data;
  formats[e->format_index].write_subtitle(e, text, start, duration);
  e->titles_written++;
  }

static void close_subtext(void * data, int do_delete)
  {
  subtext_t * e;
  e = (subtext_t *)data;
  if(e->output)
    {
    fclose(e->output);
    e->output = NULL;
    }
  if(do_delete)
    remove(e->filename);
  }

static void destroy_subtext(void * data)
  {
  subtext_t * e;
  e = (subtext_t *)data;
  if(e->output)
    close_subtext(data, 1);
  if(e->filename)
    free(e->filename);
  free(e);
  }

static bg_parameter_info_t parameters[] =
  {
    {
      name:         "format" ,
      long_name:    "Format",
      type:         BG_PARAMETER_STRINGLIST,
      val_default:  { val_str: "srt" },
      multi_names:  (char*[]){ "srt",           "mpsub",         (char*)0 },
      multi_labels: (char*[]){ "Subrip (.srt)", "MPlayer mpsub", (char*)0 },
      
    },
    { /* End of parameters */ }
  };

static bg_parameter_info_t * get_parameters_subtext(void * data)
  {
  return parameters;
  }

static void set_parameter_subtext(void * data, char * name,
                                  bg_parameter_value_t * val)
  {
  int i;
  subtext_t * e;
  e = (subtext_t *)data;

  if(!name)
    return;
  if(!strcmp(name, "format"))
    {
    for(i = 0; i < sizeof(formats)/sizeof(formats[0]); i++)
      {
      if(!strcmp(val->val_str, formats[i].name))
        {
        e->format_index = i;
        break;
        }
      }
    }
  }

bg_encoder_plugin_t the_plugin =
  {
    common:
    {
      name:           "e_subtext",       /* Unique short name */
      long_name:      "Text subtitle exporter",
      mimetypes:      NULL,
      extensions:     "srt",
      type:           BG_PLUGIN_ENCODER_SUBTITLE_TEXT,
      flags:          BG_PLUGIN_FILE,
      priority:       BG_PLUGIN_PRIORITY_MAX,
      create:         create_subtext,
      destroy:        destroy_subtext,
      get_parameters: get_parameters_subtext,
      set_parameter:  set_parameter_subtext,
      //   get_error:      get_error_subtext,
    },

    max_subtitle_text_streams: 1,
    
    get_extension:        get_extension_subtext,
    
    open:                 open_subtext,

    add_subtitle_text_stream:     add_subtitle_text_stream_subtext,
    
    start:                start_subtext,
    
    write_subtitle_text: write_subtitle_text_subtext,
    close:             close_subtext,
    
  };

/* Include this into all plugin modules exactly once
   to let the plugin loader obtain the API version */
BG_GET_PLUGIN_API_VERSION;
